#! /usr/bin/env py.test
# -*- coding: utf-8 -*-

import logging
import os

import pytest

from pypiserver import __main__, core, backend
from pypiserver.pkg_helpers import (
    normalize_pkgname_for_url,
)
from tests.doubles import Namespace


## Enable logging to detect any problems with it
##
__main__.init_logging()


def test_listdir_bad_name(tmp_path):
    tmp_path.joinpath("foo.whl").touch()
    res = list(backend.listdir(tmp_path))
    assert res == []


hashes = (
    # empty-sha256
    (
        "sha256",
        "e3b0c44298fc1c149afbf4c8996fb92427ae41e4649b934ca495991b7852b855",
    ),
    # empty-md5
    ("md5", "d41d8cd98f00b204e9800998ecf8427e"),
)


@pytest.mark.parametrize(("algo", "digest"), hashes)
def test_hashfile(tmp_path, algo, digest):
    f = tmp_path.joinpath("empty")
    f.touch()
    assert backend.digest_file(str(f), algo) == f"{algo}={digest}"


@pytest.mark.parametrize("hash_algo", ("md5", "sha256", "sha512"))
def test_fname_and_hash(tmp_path, hash_algo):
    """Ensure we are returning the expected hashes for files."""

    def digester(pkg):
        digest = backend.digest_file(pkg.fn, hash_algo)
        pkg.digest = digest
        return digest

    f = tmp_path.joinpath("tmpfile")
    f.touch()
    pkgfile = core.PkgFile("tmp", "1.0.0", str(f), f.parent, f.name)
    pkgfile.digester = digester

    assert pkgfile.fname_and_hash == f"{f.name}#{digester(pkgfile)}"


def test_redirect_project_encodes_newlines():
    """Ensure raw newlines are url encoded in the generated redirect."""
    request = Namespace(custom_fullpath="/\nSet-Cookie:malicious=1;")
    project = "\nSet-Cookie:malicious=1;"
    newpath = core.get_bad_url_redirect_path(request, project)
    assert "\n" not in newpath


def test_normalize_pkgname_for_url_encodes_newlines():
    """Ensure newlines are url encoded in package names for urls."""
    assert "\n" not in normalize_pkgname_for_url("/\nSet-Cookie:malicious=1;")



#! /usr/bin/env py.test

# Builtin imports
import logging
import os
import pathlib
import xmlrpc.client as xmlrpclib
from html import unescape

# Third party imports
import pytest
import webtest

# Local Imports
from tests.test_pkg_helpers import files, invalid_files
from pypiserver import __main__, bottle, core, Bottle
from pypiserver.backend import CachingFileBackend, SimpleFileBackend

# Enable logging to detect any problems with it
##

__main__.init_logging()


@pytest.fixture
def app(tmpdir):
    from pypiserver import app

    return app(
        roots=[pathlib.Path(tmpdir.strpath)],
        authenticate=[],
        password_file=".",
        backend_arg="simple-dir",
    )


@pytest.fixture
def testapp(app):
    """Return a webtest TestApp initiated with pypiserver app"""
    bottle.debug(True)
    return webtest.TestApp(app)


@pytest.fixture
def root(tmpdir):
    """Return a pytest temporary directory"""
    return tmpdir


@pytest.fixture
def priv(app):
    b = bottle.Bottle()
    b.mount("/priv/", app)
    return b


@pytest.fixture
def testpriv(priv):
    return webtest.TestApp(priv)


@pytest.fixture
def search_xml():
    """Return an xml dom suitable for passing to search"""
    xml = "<xml><methodName>search</methodName><string>test</string></xml>"
    return xml


@pytest.fixture(
    params=[
        "  ",  # Mustcontain test below fails when string is empty.
        "Hey there!",
        "<html><body>Hey there!</body></html>",
    ]
)
def welcome_file_no_vars(request, root):
    """Welcome file fixture

    :param request: pytest builtin fixture
    :param root: root temporary directory
    """
    wfile = root.join("testwelcome.html")
    wfile.write(request.param)

    return wfile


@pytest.fixture()
def welcome_file_all_vars(request, root):
    msg = """
    {{URL}}
    {{VERSION}}
    {{NUMPKGS}}
    {{PACKAGES}}
    {{SIMPLE}}
    """
    wfile = root.join("testwelcome.html")
    wfile.write(msg)

    return wfile


@pytest.fixture
def add_file_to_root(app):
    def file_adder(root, filename, content=""):
        root.join(filename).write(content)
        backend = app.config.backend
        if isinstance(backend, CachingFileBackend):
            backend.cache_manager.invalidate_root_cache(root)

    return file_adder


def test_root_count(root, testapp, add_file_to_root):
    """Test that the welcome page count updates with added packages

    :param root: root temporary directory fixture
    :param testapp: webtest TestApp
    """
    resp = testapp.get("/")
    resp.mustcontain("PyPI compatible package index serving 0 packages")
    add_file_to_root(root, "Twisted-11.0.0.tar.bz2")
    resp = testapp.get("/")
    resp.mustcontain("PyPI compatible package index serving 1 packages")


def test_root_hostname(testapp):
    resp = testapp.get("/", headers={"Host": "systemexit.de"})
    resp.mustcontain(
        "easy_install --index-url http://systemexit.de/simple/ PACKAGE"
    )
    # go("http://systemexit.de/")


def test_root_welcome_msg_no_vars(root, welcome_file_no_vars):
    from pypiserver import app

    app = app(root=root.strpath, welcome_file=welcome_file_no_vars.strpath)
    testapp = webtest.TestApp(app)
    resp = testapp.get("/")
    from pypiserver import __version__ as pver

    resp.mustcontain(welcome_file_no_vars.read(), no=pver)


def test_root_welcome_msg_all_vars(root, welcome_file_all_vars):
    from pypiserver import app

    app = app(root=root.strpath, welcome_file=welcome_file_all_vars.strpath)
    testapp = webtest.TestApp(app)
    resp = testapp.get("/")

    from pypiserver import __version__ as pver

    resp.mustcontain(pver)


def test_root_welcome_msg_antiXSS(testapp):
    """https://github.com/pypiserver/pypiserver/issues/77"""
    resp = testapp.get("/?<alert>Red</alert>", headers={"Host": "somehost.org"})
    resp.mustcontain("alert", "somehost.org", no="<alert>")


def test_root_remove_not_found_msg_antiXSS(testapp):
    """https://github.com/pypiserver/pypiserver/issues/77"""
    resp = testapp.post(
        "/",
        expect_errors=True,
        headers={"Host": "somehost.org"},
        params={
            ":action": "remove_pkg",
            "name": "<alert>Red</alert>",
            "version": "1.1.1",
        },
    )
    resp.mustcontain("alert", "somehost.org", no="<alert>")


def test_packages_redirect(testapp):
    resp = testapp.get("/packages")
    assert resp.status_code >= 300
    assert resp.status_code < 400
    assert resp.location.endswith("/packages/")


def test_packages_empty(testapp):
    resp = testapp.get("/packages/")
    assert len(resp.html("a")) == 0


def test_health_default_endpoint(testapp):
    resp = testapp.get("/health")
    assert resp.status_int == 200
    assert "Ok" in resp


def test_health_customized_endpoint(root):
    from pypiserver import app

    _app = app(root=root.strpath, health_endpoint="/healthz")
    testapp = webtest.TestApp(_app)
    resp = testapp.get("/healthz")
    assert resp.status_int == 200
    assert "Ok" in resp


def test_health_invalid_customized_endpoint(root):
    from pypiserver import app

    with pytest.raises(RuntimeError, match="overlaps with existing routes"):
        app(root=root.strpath, health_endpoint="/simple")


def test_favicon(testapp):
    testapp.get("/favicon.ico", status=404)


def test_fallback(testapp):
    assert not testapp.app._pypiserver_config.disable_fallback
    resp = testapp.get("/simple/pypiserver/", status=302)
    assert resp.headers["Location"] == "https://pypi.org/simple/pypiserver/"


def test_no_fallback(testapp):
    testapp.app._pypiserver_config.disable_fallback = True
    testapp.get("/simple/pypiserver/", status=404)


def test_serve_no_dotfiles(root, testapp):
    root.join(".foo-1.0.zip").write("secret")
    testapp.get("/packages/.foo-1.0.zip", status=404)


def test_packages_list_no_dotfiles(root, testapp):
    root.join(".foo-1.0.zip").write("secret")
    resp = testapp.get("/packages/")
    assert "foo" not in resp


def test_simple_redirect(testapp):
    resp = testapp.get("/simple")
    assert resp.status_code >= 300
    assert resp.status_code < 400
    assert resp.location.endswith("/simple/")


def test_simple_list_no_dotfiles(root, testapp):
    root.join(".foo-1.0.zip").write("secret")
    resp = testapp.get("/simple/")
    assert "foo" not in resp


def test_simple_list_no_dotfiles2(root, testapp):
    root.join(".foo-1.0.zip").write("secret")
    resp = testapp.get("/simple/")
    assert resp.html("a") == []


def test_serve_no_dotdir(root, testapp):
    root.mkdir(".subdir").join("foo-1.0.zip").write("secret")
    testapp.get("/packages/.subdir/foo-1.0.zip", status=404)


def test_packages_list_no_dotdir(root, testapp):
    root.mkdir(".subdir").join("foo-1.0.zip").write("secret")
    resp = testapp.get("/packages/")
    assert "foo" not in resp


def test_simple_list_no_dotdir(root, testapp):
    root.mkdir(".subdir").join("foo-1.0.zip").write("secret")
    resp = testapp.get("/simple/")
    assert "foo" not in resp


def test_simple_list_no_dotdir2(root, testapp):
    root.mkdir(".subdir").join("foo-1.0.zip").write("secret")
    resp = testapp.get("/simple/foo/")
    assert resp.html("a") == []


def test_simple_name_redirect(testapp):
    resp = testapp.get("/simple/foobar")
    assert resp.status_code >= 300
    assert resp.status_code < 400
    assert resp.location.endswith("/simple/foobar/")


@pytest.mark.parametrize(
    "package,normalized",
    [
        ("FooBar", "foobar"),
        ("Foo.Bar", "foo-bar"),
        ("foo_bar", "foo-bar"),
        ("Foo-Bar", "foo-bar"),
        ("foo--_.bar", "foo-bar"),
    ],
)
def test_simple_normalized_name_redirect(testapp, package, normalized):
    resp = testapp.get("/simple/{0}/".format(package))
    assert resp.status_code >= 300
    assert resp.status_code < 400
    assert resp.location.endswith("/simple/{0}/".format(normalized))


def test_simple_index(root, testapp):
    root.join("foobar-1.0.zip").write("")
    root.join("foobar-1.1.zip").write("")
    root.join("foobarbaz-1.1.zip").write("")
    root.join("foobar.baz-1.1.zip").write("")

    resp = testapp.get("/simple/foobar/")
    assert len(resp.html("a")) == 2


def test_simple_index_list(root, testapp):
    root.join("foobar-1.0.zip").write("")
    root.join("foobar-1.1.zip").write("")
    root.join("foobarbaz-1.1.zip").write("")
    root.join("foobar.baz-1.1.zip").write("")

    resp = testapp.get("/simple/")
    assert len(resp.html("a")) == 3


def test_simple_index_case(root, testapp):
    root.join("FooBar-1.0.zip").write("")
    root.join("FooBar-1.1.zip").write("")
    resp = testapp.get("/simple/foobar/")
    assert len(resp.html("a")) == 2


def test_nonroot_root(testpriv):
    resp = testpriv.get("/priv/", headers={"Host": "nonroot"})
    resp.mustcontain(
        "easy_install --index-url http://nonroot/priv/simple/ PACKAGE"
    )


def test_nonroot_root_with_x_forwarded_host(testapp):
    resp = testapp.get("/", headers={"X-Forwarded-Host": "forward.ed/priv/"})
    resp.mustcontain(
        "easy_install --index-url http://forward.ed/priv/simple/ PACKAGE"
    )
    resp.mustcontain("""<a href="/priv/packages/">here</a>""")


def test_nonroot_root_with_x_forwarded_host_without_trailing_slash(testapp):
    resp = testapp.get("/", headers={"X-Forwarded-Host": "forward.ed/priv"})
    resp.mustcontain(
        "easy_install --index-url http://forward.ed/priv/simple/ PACKAGE"
    )
    resp.mustcontain("""<a href="/priv/packages/">here</a>""")


def test_nonroot_simple_index(root, testpriv, add_file_to_root):
    add_file_to_root(root, "foobar-1.0.zip", "123")
    resp = testpriv.get("/priv/simple/foobar/")
    links = resp.html("a")
    assert len(links) == 1
    assert links[0]["href"].startswith("/priv/packages/foobar-1.0.zip#")


def test_nonroot_simple_index_with_x_forwarded_host(
    root, testapp, add_file_to_root
):
    add_file_to_root(root, "foobar-1.0.zip", "123")

    resp = testapp.get(
        "/simple/foobar/", headers={"X-Forwarded-Host": "forwarded.ed/priv/"}
    )
    links = resp.html("a")
    assert len(links) == 1
    assert links[0]["href"].startswith("/priv/packages/foobar-1.0.zip#")


def test_nonroot_simple_packages(root, testpriv, add_file_to_root):
    add_file_to_root(root, "foobar-1.0.zip", "123")
    resp = testpriv.get("/priv/packages/")
    links = resp.html("a")
    assert len(links) == 1
    assert "/priv/packages/foobar-1.0.zip#" in links[0]["href"]


def test_nonroot_simple_packages_with_x_forwarded_host(
    root, testapp, add_file_to_root
):
    add_file_to_root(root, "foobar-1.0.zip", "123")

    resp = testapp.get(
        "/packages/", headers={"X-Forwarded-Host": "forwarded/priv/"}
    )
    links = resp.html("a")
    assert len(links) == 1
    assert "/priv/packages/foobar-1.0.zip#" in links[0]["href"]


def test_root_no_relative_paths(testpriv):
    """https://github.com/pypiserver/pypiserver/issues/25"""
    resp = testpriv.get("/priv/")
    hrefs = [x["href"] for x in resp.html("a")]
    assert hrefs == [
        "/priv/packages/",
        "/priv/simple/",
        "https://pypi.org/project/pypiserver/",
    ]


def test_simple_index_list_no_duplicates(root, testapp):
    root.join("foo-bar-1.0.tar.gz").write("")
    root.join("foo_bar-1.0-py2.7.egg").write("")

    resp = testapp.get("/simple/")
    assert len(resp.html("a")) == 1


def test_simple_index_list_name_with_underscore(root, testapp):
    root.join("foo_bar-1.0.tar.gz").write("")
    root.join("foo_bar-1.0-py2.7.egg").write("")

    resp = testapp.get("/simple/")
    assert len(resp.html("a")) == 1
    hrefs = [x["href"] for x in resp.html("a")]
    assert hrefs == ["foo-bar/"]


def test_simple_index_egg_and_tarball(root, testapp):
    root.join("foo-bar-1.0.tar.gz").write("")
    root.join("foo_bar-1.0-py2.7.egg").write("")

    resp = testapp.get("/simple/foo-bar/")
    assert len(resp.html("a")) == 2


def test_simple_index_list_name_with_underscore_no_egg(root, testapp):
    root.join("foo_bar-1.0.tar.gz").write("")
    root.join("foo-bar-1.1.tar.gz").write("")

    resp = testapp.get("/simple/")
    assert len(resp.html("a")) == 1
    hrefs = set([x["href"] for x in resp.html("a")])
    assert hrefs == {"foo-bar/"}


def test_json_info(root, testapp):
    root.join("foobar-1.0.zip").write("")
    root.join("foobar-1.1.zip").write("")
    root.join("foobar-1.1-linux.zip").write("")
    root.join("foobarX-1.1.zip").write("")

    resp = testapp.get("/foobar/json")
    assert "info" in resp.json
    assert "releases" in resp.json
    assert len(resp.json["info"]) == 1
    assert len(resp.json["releases"]) == 2
    assert len(resp.json["releases"]["1.0"]) == 1
    assert len(resp.json["releases"]["1.1"]) == 2


def test_json_info_package_not_existing(root, testapp):
    resp = testapp.get("/foobar/json", status=404)


@pytest.mark.parametrize(
    "package,normalized",
    [
        ("FooBar", "foobar"),
        ("Foo.Bar", "foo-bar"),
        ("foo_bar", "foo-bar"),
        ("Foo-Bar", "foo-bar"),
        ("foo--_.bar", "foo-bar"),
    ],
)
def test_json_info_normalized_name_redirect(testapp, package, normalized):
    resp = testapp.get("/{0}/json".format(package))
    assert resp.status_code >= 300
    assert resp.status_code < 400
    assert resp.location.endswith("/{0}/json".format(normalized))


def test_no_cache_control_set(root, testapp):
    assert not testapp.app._pypiserver_config.cache_control
    root.join("foo_bar-1.0.tar.gz").write("")
    resp = testapp.get("/packages/foo_bar-1.0.tar.gz")
    assert "Cache-Control" not in resp.headers


def test_cache_control_set(root):
    from pypiserver import app

    AGE = 86400
    app_with_cache = webtest.TestApp(app(root=root.strpath, cache_control=AGE))
    root.join("foo_bar-1.0.tar.gz").write("")
    resp = app_with_cache.get("/packages/foo_bar-1.0.tar.gz")
    assert "Cache-Control" in resp.headers
    assert resp.headers["Cache-Control"] == f"public, max-age={AGE}"


def test_upload_noAction(testapp):
    resp = testapp.post("/", expect_errors=1)
    assert resp.status == "400 Bad Request"
    assert "Missing ':action' field!" in unescape(resp.text)


def test_upload_badAction(testapp):
    resp = testapp.post("/", params={":action": "BAD"}, expect_errors=1)
    assert resp.status == "400 Bad Request"
    assert "Unsupported ':action' field: BAD" in unescape(resp.text)


@pytest.mark.parametrize(
    "package", [f[0] for f in files if f[1] and "/" not in f[0]]
)
def test_upload(package, root, testapp):
    resp = testapp.post(
        "/",
        params={":action": "file_upload"},
        upload_files=[("content", package, b"")],
    )
    assert resp.status_int == 200
    uploaded_pkgs = [f.basename for f in root.listdir()]
    assert len(uploaded_pkgs) == 1
    assert uploaded_pkgs[0].lower() == package.lower()


def test_upload_conflict_on_existing(root, testapp):
    package = "foo_bar-1.0.tar.gz"
    root.join("foo_bar-1.0.tar.gz").write("")

    resp = testapp.post(
        "/",
        params={":action": "file_upload"},
        upload_files=[("content", package, b"")],
        status=409,
    )

    assert resp.status_int == 409
    assert "Package 'foo_bar-1.0.tar.gz' already exists!" in unescape(resp.text)


@pytest.mark.parametrize(
    "package", [f[0] for f in files if f[1] and "/" not in f[0]]
)
def test_upload_with_signature(package, root, testapp):
    resp = testapp.post(
        "/",
        params={":action": "file_upload"},
        upload_files=[
            ("content", package, b""),
            ("gpg_signature", f"{package}.asc", b""),
        ],
    )
    assert resp.status_int == 200
    uploaded_pkgs = [f.basename.lower() for f in root.listdir()]
    assert len(uploaded_pkgs) == 2
    assert package.lower() in uploaded_pkgs
    assert f"{package.lower()}.asc" in uploaded_pkgs


@pytest.mark.parametrize("package", invalid_files)
def test_upload_badFilename(package, root, testapp):
    resp = testapp.post(
        "/",
        params={":action": "file_upload"},
        upload_files=[("content", package, b"")],
        expect_errors=1,
    )
    assert resp.status == "400 Bad Request"
    assert f"Bad filename: {package}" in resp.text


@pytest.mark.parametrize(
    "pkgs,matches",
    [
        ([], []),
        (["test-1.0.tar.gz"], [("test", "1.0")]),
        (
            ["test-1.0.tar.gz", "test-test-2.0.1.tar.gz"],
            [("test", "1.0"), ("test-test", "2.0.1")],
        ),
        (["test-1.0.tar.gz", "other-2.0.tar.gz"], [("test", "1.0")]),
        (["test-2.0-py2.py3-none-any.whl"], [("test", "2.0")]),
        (["other-2.0.tar.gz"], []),
    ],
)
def test_search(root, testapp, search_xml, pkgs, matches):
    """Test the search functionality at the RPC2 endpoint

    Calls the handle_rpc function by posting to the WebTest server with
    the string returned by the ``search_xml`` fixture as the request
    body. The result is parsed for convenience by the xmlrpc.client
    (xmlrpclib in Python 2.x). The parsed result is a 2-tuple. The
    second item is the method called, in this case always "search". The
    first item is a 1-tuple which contains a list of match information
    as dicts, e.g:

    ``(([{'version': '2.0', '_pypi_ordering': 0,
    'name': 'test', 'summary': '2.0'}],), 'search')``

    The pkgs parameter is a list of items to write to the packages
    directory, which should then be available for the subsequent
    search. The matches parameter is a list of 2-tuples of the form
    (``name``, ``version``), where ``name`` and ``version`` are the
    expected name and version matches for a search for the "test"
    package as specified by the search_xml fixture.

    :param root: root temporary directory fixture; used as packages dir
        for testapp
    :param testapp: webtest TestApp
    :param str search_xml: XML string roughly equivalent to a pip search
        for "test"
    :param pkgs: package file names to be written into packages
        directory
    :param matches: a list of 2-tuples containing expected (name,
        version) matches for the "test" query
    """
    for pkg in pkgs:
        root.join(pkg).write("")
    resp = testapp.post("/RPC2", search_xml)
    parsed = xmlrpclib.loads(resp.text)
    assert len(parsed) == 2 and parsed[1] == "search"
    if not matches:
        assert len(parsed[0]) == 1 and not parsed[0][0]
    else:
        assert len(parsed[0][0]) == len(matches) and parsed[0][0]
        for returned in parsed[0][0]:
            print(returned)
            assert returned["name"] in [match[0] for match in matches]
            assert returned["version"] in [match[1] for match in matches]


class TestRemovePkg:
    """The API allows removal of packages."""

    @pytest.mark.parametrize(
        "pkg, name, ver",
        (
            ("test-1.0.tar.gz", "test", "1.0"),
            ("test-1.0-py2-py3-none-any.whl", "test", "1.0"),
        ),
    )
    def test_remove_pkg(self, root, testapp, pkg, name, ver):
        """Packages can be removed via POST."""
        root.join(pkg).write("")
        assert len(os.listdir(str(root))) == 1
        params = {":action": "remove_pkg", "name": name, "version": ver}
        testapp.post("/", params=params)
        assert len(os.listdir(str(root))) == 0

    @pytest.mark.parametrize(
        "pkg, name, ver",
        (
            ("test-1.0.tar.gz", "test", "1.0"),
            ("test-1.0-py2-py3-none-any.whl", "test", "1.0"),
        ),
    )
    @pytest.mark.parametrize(
        "other",
        (
            "test-2.0.tar.gz",
            "test-2.0-py2-py3-none-any.whl",
            "other-1.0.tar.gz",
            "other-1.0-py2-py3-none-any.whl",
        ),
    )
    def test_remove_pkg_only_targeted(
        self, root, testapp, pkg, name, ver, other
    ):
        """Only the targeted package is removed."""
        root.join(pkg).write("")
        root.join(other).write("")
        assert len(os.listdir(str(root))) == 2
        params = {":action": "remove_pkg", "name": name, "version": ver}
        testapp.post("/", params=params)
        assert len(os.listdir(str(root))) == 1

    @pytest.mark.parametrize(
        "pkgs, name, ver",
        (
            (
                ("test-1.0.tar.gz", "test-1.0-py2-py3-none-any.whl"),
                "test",
                "1.0",
            ),
        ),
    )
    def test_all_instances_removed(self, root, testapp, pkgs, name, ver):
        """Test that all instances of the target are removed."""
        for pkg in pkgs:
            root.join(pkg).write("")
        assert len(os.listdir(str(root))) == len(pkgs)
        params = {":action": "remove_pkg", "name": name, "version": ver}
        testapp.post("/", params=params)
        assert len(os.listdir(str(root))) == 0

    @pytest.mark.parametrize(
        ("name", "version"),
        [
            (None, None),
            (None, ""),
            ("", None),
            (None, "1"),
            ("pkg", None),
            ("", "1"),
            ("pkg", ""),
        ],
    )
    def test_remove_pkg_missingNaveVersion(self, name, version, root, testapp):
        msg = f"Missing 'name'/'version' fields: name={name}, version={version}"
        params = {":action": "remove_pkg", "name": name, "version": version}
        params = dict((k, v) for k, v in params.items() if v is not None)
        resp = testapp.post("/", expect_errors=1, params=params)

        assert resp.status == "400 Bad Request"
        assert msg in unescape(resp.text)

    def test_remove_pkg_notFound(self, root, testapp):
        resp = testapp.post(
            "/",
            expect_errors=1,
            params={":action": "remove_pkg", "name": "foo", "version": "123"},
        )
        assert resp.status == "404 Not Found"
        assert "foo (123) not found" in unescape(resp.text)



#!/usr/bin/env py.test
"""Tests for manage.py."""

from __future__ import absolute_import, print_function, unicode_literals

from pathlib import Path
from unittest.mock import Mock

import py
import pytest

from pypiserver import manage
from pypiserver.core import PkgFile
from pypiserver.pkg_helpers import guess_pkgname_and_version, parse_version
from pypiserver.manage import (
    PipCmd,
    build_releases,
    filter_stable_releases,
    filter_latest_pkgs,
    is_stable_version,
    update_package,
    update_all_packages,
)


def touch_files(root, files):
    root = py.path.local(root)  # pylint: disable=no-member
    for f in files:
        root.join(f).ensure()


def pkgfile_from_path(fn):
    pkgname, version = guess_pkgname_and_version(fn)
    return PkgFile(
        pkgname=pkgname,
        version=version,
        root=py.path.local(fn)
        .parts()[1]
        .strpath,  # noqa pylint: disable=no-member
        fn=fn,
    )


@pytest.mark.parametrize(
    ("version", "is_stable"),
    [
        ("1.0", True),
        ("0.0.0", True),
        ("1.1beta1", False),
        ("1.2.10-123", True),
        ("5.5.0-DEV", False),
        ("1.2-rc1", False),
        ("1.0b1", False),
    ],
)
def test_is_stable_version(version, is_stable):
    parsed_version = parse_version(version)
    assert is_stable_version(parsed_version) == is_stable


def test_build_releases():
    p = pkgfile_from_path("/home/ralf/pypiserver/d/greenlet-0.2.zip")

    expected = dict(
        parsed_version=("00000000", "00000003", "*final"),
        pkgname="greenlet",
        replaces=p,
        version="0.3.0",
    )

    (res,) = list(build_releases(p, ["0.3.0"]))
    for k, v in expected.items():
        assert getattr(res, k) == v


def test_filter_stable_releases():
    p = pkgfile_from_path("/home/ralf/pypiserver/d/greenlet-0.2.zip")
    assert list(filter_stable_releases([p])) == [p]

    p2 = pkgfile_from_path("/home/ralf/pypiserver/d/greenlet-0.5rc1.zip")
    assert list(filter_stable_releases([p2])) == []


def test_filter_latest_pkgs():
    paths = [
        "/home/ralf/greenlet-0.2.zip",
        "/home/ralf/foo/baz-1.0.zip" "/home/ralf/bar/greenlet-0.3.zip",
    ]
    pkgs = [pkgfile_from_path(x) for x in paths]

    assert frozenset(filter_latest_pkgs(pkgs)) == frozenset(pkgs[1:])


def test_filter_latest_pkgs_case_insensitive():
    paths = [
        "/home/ralf/greenlet-0.2.zip",
        "/home/ralf/foo/baz-1.0.zip" "/home/ralf/bar/Greenlet-0.3.zip",
    ]
    pkgs = [pkgfile_from_path(x) for x in paths]

    assert frozenset(filter_latest_pkgs(pkgs)) == frozenset(pkgs[1:])


@pytest.mark.parametrize(
    "pip_ver, cmd_type",
    (
        ("10.0.0", "d"),
        ("10.0.0rc10", "d"),
        ("10.0.0b10", "d"),
        ("10.0.0a3", "d"),
        ("10.0.0.dev8", "d"),
        ("10.0.0.dev8", "d"),
        ("18.0", "d"),
        ("9.9.8", "i"),
        ("9.9.8rc10", "i"),
        ("9.9.8b10", "i"),
        ("9.9.8a10", "i"),
        ("9.9.8.dev10", "i"),
        ("9.9", "i"),
    ),
)
def test_pip_cmd_root(pip_ver, cmd_type):
    """Verify correct determination of the command root by pip version."""
    exp_cmd = (
        "pip",
        "-q",
        "install" if cmd_type == "i" else "download",
    )
    assert tuple(PipCmd.update_root(pip_ver)) == exp_cmd


def test_pip_cmd_update():
    """Verify the correct determination of a pip command."""
    index = "https://pypi.org/simple"
    destdir = "foo/bar"
    pkg_name = "mypkg"
    pkg_version = "12.0"
    cmd_root = ("pip", "-q", "download")
    exp_cmd = cmd_root + (
        "--no-deps",
        "-i",
        index,
        "-d",
        destdir,
        "{}=={}".format(pkg_name, pkg_version),
    )
    assert exp_cmd == tuple(
        PipCmd.update(cmd_root, destdir, pkg_name, pkg_version)
    )


def test_pip_cmd_update_index_overridden():
    """Verify the correct determination of a pip command."""
    index = "https://pypi.org/complex"
    destdir = "foo/bar"
    pkg_name = "mypkg"
    pkg_version = "12.0"
    cmd_root = ("pip", "-q", "download")
    exp_cmd = cmd_root + (
        "--no-deps",
        "-i",
        index,
        "-d",
        destdir,
        "{}=={}".format(pkg_name, pkg_version),
    )
    assert exp_cmd == tuple(
        PipCmd.update(cmd_root, destdir, pkg_name, pkg_version, index=index)
    )


def test_update_package(monkeypatch):
    """Test generating an update command for a package."""
    monkeypatch.setattr(manage, "call", Mock())
    pkg = PkgFile("mypkg", "1.0", replaces=PkgFile("mypkg", "0.9"))
    update_package(pkg, ".")
    manage.call.assert_called_once_with(
        (  # pylint: disable=no-member
            "pip",
            "-q",
            "download",
            "--no-deps",
            "-i",
            "https://pypi.org/simple",
            "-d",
            ".",
            "mypkg==1.0",
        )
    )


def test_update_package_dry_run(monkeypatch):
    """Test generating an update command for a package."""
    monkeypatch.setattr(manage, "call", Mock())
    pkg = PkgFile("mypkg", "1.0", replaces=PkgFile("mypkg", "0.9"))
    update_package(pkg, ".", dry_run=True)
    assert not manage.call.mock_calls  # pylint: disable=no-member


def test_update_all_packages(monkeypatch):
    """Test calling update_all_packages()"""
    public_pkg_1 = PkgFile("Flask", "1.0")
    public_pkg_2 = PkgFile("requests", "1.0")
    private_pkg_1 = PkgFile("my_private_pkg", "1.0")
    private_pkg_2 = PkgFile("my_other_private_pkg", "1.0")

    roots_mock = {
        Path("/opt/pypi"): [
            public_pkg_1,
            private_pkg_1,
        ],
        Path("/data/pypi"): [public_pkg_2, private_pkg_2],
    }

    def core_listdir_mock(path: Path):
        return roots_mock.get(path, [])

    monkeypatch.setattr(manage, "listdir", core_listdir_mock)
    monkeypatch.setattr(manage, "update", Mock(return_value=None))

    destdir = None
    dry_run = False
    stable_only = True

    update_all_packages(
        roots=list(roots_mock.keys()),
        destdir=destdir,
        dry_run=dry_run,
        stable_only=stable_only,
        ignorelist=None,
    )

    manage.update.assert_called_once_with(  # pylint: disable=no-member
        frozenset([public_pkg_1, public_pkg_2, private_pkg_1, private_pkg_2]),
        destdir,
        dry_run,
        stable_only,
    )


def test_update_all_packages_with_blacklist(monkeypatch):
    """Test calling update_all_packages()"""
    public_pkg_1 = PkgFile("Flask", "1.0")
    public_pkg_2 = PkgFile("requests", "1.0")
    private_pkg_1 = PkgFile("my_private_pkg", "1.0")
    private_pkg_2 = PkgFile("my_other_private_pkg", "1.0")

    roots_mock = {
        Path("/opt/pypi"): [
            public_pkg_1,
            private_pkg_1,
        ],
        Path("/data/pypi"): [public_pkg_2, private_pkg_2],
    }

    def core_listdir_mock(path: Path):
        return roots_mock.get(path, [])

    monkeypatch.setattr(manage, "listdir", core_listdir_mock)
    monkeypatch.setattr(manage, "update", Mock(return_value=None))

    destdir = None
    dry_run = False
    stable_only = True

    update_all_packages(
        roots=list(roots_mock.keys()),
        destdir=destdir,
        dry_run=dry_run,
        stable_only=stable_only,
        ignorelist=["my_private_pkg", "my_other_private_pkg"],
    )

    manage.update.assert_called_once_with(  # pylint: disable=no-member
        frozenset([public_pkg_1, public_pkg_2]), destdir, dry_run, stable_only
    )



#!/usr/bin/env python3

"""Output expected docker tags to build in CI."""

import json
import typing as t
import re
from argparse import ArgumentParser, Namespace


RELEASE_RE = re.compile(r"v[0-9]+\.[0-9]+\.[0-9]+(\.post[0-9]+)?")
PRE_RELEASE_RE = re.compile(r"v[0-9]+\.[0-9]+\.[0-9]+(a|b|c|\.?dev)[0-9]+")


def parse_args() -> Namespace:
    """Parse cmdline args."""
    parser = ArgumentParser()
    parser.add_argument(
        "ref",
        help=(
            "The github ref for which CI is running. This may be a full ref "
            "like refs/tags/v1.2.3 or refs/heads/master, or just a tag/branch "
            "name like v1.2.3 or master."
        ),
    )
    parser.add_argument(
        "action",
        help=("The action to perform"),
        choices=("docker_tags", "pypi_release", "has_tags"),
    )
    return parser.parse_args()


def strip_ref_to_name(ref: str) -> str:
    """Strip a full ref to a name."""
    strips = ("refs/heads/", "refs/tags/")
    for strip in strips:
        if ref.startswith(strip):
            return ref[len(strip) :]
    return ref


def name_to_array(name: str) -> t.Tuple[str, ...]:
    """Convert a ref name to an array of tags to build."""
    tags: t.Dict[str, t.Callable[[str], bool]] = {
        # unstable for any master build
        "unstable": lambda i: i == "master",
        # latest goes for full releases
        "latest": lambda i: RELEASE_RE.fullmatch(i) is not None,
        # the tag itself for any release or pre-release tag
        name: lambda i: (
            RELEASE_RE.fullmatch(i) is not None
            or PRE_RELEASE_RE.fullmatch(i) is not None
        ),
    }
    return tuple(tag for tag, test in tags.items() if test(name))


def ref_to_json(ref: str) -> str:
    """Convert a ref to a JSON array and return it as a string."""
    array = name_to_array(strip_ref_to_name(ref))
    return json.dumps(array)


def should_deploy_to_pypi(ref: str) -> str:
    """Return a JSON bool indicating whether we should deploy to PyPI."""
    name = strip_ref_to_name(ref)
    return json.dumps(
        RELEASE_RE.fullmatch(name) is not None
        or PRE_RELEASE_RE.fullmatch(name) is not None
    )


def main() -> None:
    """Parse args and print the JSON array."""
    args = parse_args()
    action_switch: t.Dict[str, t.Callable[[], None]] = {
        "docker_tags": lambda: print(ref_to_json(args.ref)),
        "has_tags": lambda: print(
            json.dumps(len(name_to_array(strip_ref_to_name(args.ref))) > 0)
        ),
        "pypi_release": lambda: print(should_deploy_to_pypi(args.ref)),
    }
    action_switch[args.action]()


if __name__ == "__main__":
    main()



#! /usr/bin/env python3

import re
from pathlib import Path

from setuptools import setup

tests_require = [
    "pytest>=2.3",
    "tox",
    "twine",
    "passlib>=1.6",
    "webtest",
    "build>=1.2.0;python_version>='3.8'",
]

setup_requires = [
    "setuptools",
    "setuptools-git>=0.3",
    "wheel>=0.25.0",
]
install_requires = [
    "pip>=7",
    "packaging>=23.2",
    "importlib_resources;python_version>'3.8' and python_version<'3.12'",
]


def read_file(rel_path: str):
    return Path(__file__).parent.joinpath(rel_path).read_text()


def get_version():
    locals_ = {}
    version_line = re.compile(
        r'^[\w =]*__version__ = "\d+\.\d+\.\d+\.?\w*\d*"$'
    )
    try:
        for ln in filter(
            version_line.match,
            read_file("pypiserver/__init__.py").splitlines(),
        ):
            exec(ln, locals_)
    except (ImportError, RuntimeError):
        pass
    return locals_["__version__"]


setup(
    name="pypiserver",
    description="A minimal PyPI server for use with pip/easy_install.",
    long_description=read_file("README.md"),
    long_description_content_type="text/markdown",
    version=get_version(),
    packages=["pypiserver"],
    package_data={"pypiserver": ["welcome.html"]},
    python_requires=">=3.6",
    install_requires=install_requires,
    setup_requires=setup_requires,
    extras_require={"passlib": ["passlib>=1.6"], "cache": ["watchdog"]},
    tests_require=tests_require,
    url="https://github.com/pypiserver/pypiserver",
    maintainer=(
        "Kostis Anagnostopoulos <ankostis@gmail.com>"
        "Matthew Planchard <mplanchard@gmail.com>"
    ),
    maintainer_email="ankostis@gmail.com",
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Environment :: Web Environment",
        "Intended Audience :: Developers",
        "Intended Audience :: System Administrators",
        "License :: OSI Approved :: BSD License",
        "License :: OSI Approved :: zlib/libpng License",
        "Operating System :: MacOS :: MacOS X",
        "Operating System :: Microsoft :: Windows",
        "Operating System :: POSIX",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: Implementation :: CPython",
        "Programming Language :: Python :: Implementation :: PyPy",
        "Topic :: Software Development :: Build Tools",
        "Topic :: System :: Software Distribution",
    ],
    zip_safe=True,
    entry_points={
        "paste.app_factory": ["main=pypiserver:paste_app_factory"],
        "console_scripts": ["pypi-server=pypiserver.__main__:main"],
    },
    options={"bdist_wheel": {"universal": True}},
    platforms=["any"],
)



#! /usr/bin/env py.test
from pathlib import Path

import pytest
import re
from pypiserver import version as my_ver


@pytest.fixture()
def readme():
    return Path(__file__).parents[1].joinpath("README.md").read_text()


def test_READMEversion(readme):
    m = re.compile(
        r"^\|\s*Version\s*\|\s*(\d+\.\d+\.\d+)\s*\|$", re.MULTILINE
    ).search(readme)
    assert m, "Could not find version on README!"
    assert m.group(1) == my_ver, f"Updated version({m.group(1)}) on README!"



"""Tests for the Pypiserver Docker image."""

import contextlib
import os
import shutil
import socket
import subprocess
import sys
import tempfile
import time
import typing as t
from pathlib import Path

import httpx
import pypiserver
import pytest


PYPISERVER_PROCESS_NAME = "pypi-server"
TEST_DEMO_PIP_PACKAGE = "pypiserver-mypkg"

THIS_DIR = Path(__file__).parent
ROOT_DIR = THIS_DIR.parent
DOCKERFILE = ROOT_DIR / "Dockerfile"
FIXTURES = ROOT_DIR / "fixtures"
MYPKG_ROOT = FIXTURES / "mypkg"
HTPASS_FILE = FIXTURES / "htpasswd.a.a"


# This rule is largely useless when using pytest because of the need to use the
# name of the fixture as an argument to the test function or fixture using it
# pylint: disable=redefined-outer-name
#
# Also useless rule for our test context, where we may want to group test
# functions in a class to share common fixtures, but where we don't care about
# the `self` instance.
# pylint: disable=no-self-use


@pytest.fixture(scope="session")
def image() -> str:
    """Build the docker image for pypiserver.

    Return the tag.
    """
    tag = "pypiserver:test"
    run(
        "docker",
        "build",
        "--file",
        str(DOCKERFILE),
        "--tag",
        tag,
        str(ROOT_DIR),
        cwd=ROOT_DIR,
    )
    return tag


@pytest.fixture(scope="session")
def mypkg_build() -> None:
    """Ensure the mypkg test fixture package is build."""
    # Use make for this so that it will skip the build step if it's not needed
    run("make", "mypkg", cwd=ROOT_DIR)


@pytest.fixture(scope="session")
def mypkg_paths(
    mypkg_build: None,  # pylint: disable=unused-argument
) -> t.Dict[str, Path]:
    """The path to the mypkg sdist file."""
    dist_dir = Path(MYPKG_ROOT) / "dist"
    assert dist_dir.exists()

    sdist = dist_dir / "pypiserver_mypkg-1.0.0.tar.gz"
    assert sdist.exists()

    wheel = dist_dir / "pypiserver_mypkg-1.0.0-py2.py3-none-any.whl"
    assert wheel.exists()

    return {
        "dist_dir": dist_dir,
        "sdist": sdist,
        "wheel": wheel,
    }


def wait_for_container(port: int) -> None:
    """Wait for the container to be available."""
    for _ in range(60):
        try:
            httpx.get(f"http://localhost:{port}").raise_for_status()
        except (httpx.RequestError, httpx.HTTPStatusError):
            time.sleep(1)
        else:
            return

    # If we reach here, we've tried 60 times without success, meaning either
    # the container is broken or it took more than about a minute to become
    # functional, either of which cases is something we will want to look into.
    raise RuntimeError("Could not connect to pypiserver container")


def get_socket() -> int:
    """Find a random, open socket and return it."""
    # Close the socket automatically upon exiting the block
    with contextlib.closing(
        socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    ) as sock:
        # Bind to a random open socket >=1024
        sock.bind(("", 0))
        # Return the socket number
        return sock.getsockname()[1]


class RunReturn(t.NamedTuple):
    """Simple wrapper around a simple subprocess call's results."""

    returncode: int
    out: str
    err: str


def run(
    *cmd: str,
    capture: bool = False,
    raise_on_err: bool = True,
    check_code: t.Callable[[int], bool] = lambda c: c == 0,
    **popen_kwargs: t.Any,
) -> RunReturn:
    """Run a command to completion."""
    stdout = subprocess.PIPE if capture else None
    stderr = subprocess.PIPE if capture else None
    proc = subprocess.Popen(cmd, stdout=stdout, stderr=stderr, **popen_kwargs)
    out, err = proc.communicate()
    result = RunReturn(
        proc.returncode,
        "" if out is None else out.decode(),
        "" if err is None else err.decode(),
    )
    if raise_on_err and not check_code(result.returncode):
        raise RuntimeError(result)
    return result


def uninstall_pkgs() -> None:
    """Uninstall any packages we've installed."""
    res = run("pip", "freeze", capture=True)
    if any(
        ln.strip().startswith(TEST_DEMO_PIP_PACKAGE)
        for ln in res.out.splitlines()
    ):
        run("pip", "uninstall", "-y", TEST_DEMO_PIP_PACKAGE)


@pytest.fixture(scope="session", autouse=True)
def session_cleanup() -> t.Iterator[None]:
    """Deal with any pollution of the local env."""
    yield
    uninstall_pkgs()


@pytest.fixture()
def cleanup() -> t.Iterator[None]:
    """Clean up after tests that may have affected the env."""
    yield
    uninstall_pkgs()


class TestCommands:
    """Test commands other than `run`."""

    def test_help(self, image: str) -> None:
        """We can get help from the docker container."""
        res = run("docker", "run", image, "--help", capture=True)
        assert PYPISERVER_PROCESS_NAME in res.out

    def test_version(self, image: str) -> None:
        """We can get the version from the docker container."""
        res = run("docker", "run", image, "--version", capture=True)
        assert res.out.strip() == pypiserver.__version__


class TestPermissions:
    """Test permission validation, especially with mounted volumes."""

    @pytest.mark.parametrize("perms", (0o706, 0o701, 0o704))
    def test_needs_rx_on_data(self, image: str, perms: int) -> None:
        """Read and execute permissions are required on /data."""
        # Note we can't run this one as root because then we have to make a file
        # that even we can't delete.
        with tempfile.TemporaryDirectory() as tmpdir:
            # Make sure the directory is not readable for anyone other than
            # the owner
            data_dir = Path(tmpdir) / "data"
            data_dir.mkdir(mode=perms)

            res = run(
                "docker",
                "run",
                "--rm",
                "--user",
                # Run as a not us user ID, so access to /data will be
                # determined by the "all other users" setting
                str(os.getuid() + 1),
                "-v",
                # Mount the temporary directory as the /data directory
                f"{data_dir}:/data",
                image,
                capture=True,
                # This should error out, so we check that the code is non-zero
                check_code=lambda c: c != 0,
            )

            assert "must have read/execute access" in res.err

    @pytest.mark.parametrize(
        "extra_args",
        (("--user", str(os.getuid())), ("--user", str(os.getuid() + 1))),
    )
    def test_needs_rwx_on_packages(self, image: str, extra_args: tuple) -> None:
        """RWX permission is required on /data/packages."""
        with tempfile.TemporaryDirectory() as tmpdir:
            td_path = Path(tmpdir)
            # Make the /data directory read/writable by anyone
            td_path.chmod(0o777)
            # Add the /data/packages directory, and make it readable by anyone,
            # but writable only by the owner
            (td_path / "packages").mkdir(mode=0o444)

            res = run(
                "docker",
                "run",
                "--rm",
                *extra_args,
                "-v",
                # Mount the temporary directory as the /data directory
                f"{tmpdir}:/data",
                image,
                capture=True,
                # We should error out in this case
                check_code=lambda c: c != 0,
            )
            assert "must have read/write/execute access" in res.err

    def test_runs_as_pypiserver_user(self, image: str) -> None:
        """Test that the default run uses the pypiserver user."""
        host_port = get_socket()
        res = run(
            "docker",
            "run",
            "--rm",
            "--detach",
            "--publish",
            f"{host_port}:8080",
            image,
            capture=True,
        )
        container_id = res.out.strip()
        try:
            wait_for_container(host_port)
            res = run(
                "docker",
                "container",
                "exec",
                container_id,
                "ps",
                "a",
                capture=True,
            )
            proc_line = next(
                filter(
                    # grab the process line for the pypi-server process
                    lambda ln: PYPISERVER_PROCESS_NAME in ln,
                    res.out.splitlines(),
                )
            )
            user = proc_line.split()[1]
            # the ps command on these alpine containers doesn't always show the
            # full user name, so we only check for the first bit
            assert user.startswith("pypi")
        finally:
            run("docker", "container", "rm", "-f", container_id)


class ContainerInfo(t.NamedTuple):
    """Info about a running container"""

    container_id: str
    port: int
    args: tuple


class TestBasics:
    """Test basic pypiserver functionality in a simple unauthed container."""

    # We want to automatically parametrize this class' tests with a variety of
    # pypiserver args, since it should work the same in all of these cases
    @pytest.fixture(
        scope="class",
        params=[
            # default (gunicorn) server with cached backend
            (),
            # default (gunicorn) server with non-cached backend
            ("--backend", "simple-dir"),
            # explicit gunicorn server with a non-cached backend
            ("--server", "gunicorn", "--backend", "simple-dir"),
            # explicit gunicorn server
            ("--server", "gunicorn"),
            # explicit waitress server
            ("--server", "wsgiref"),
            # explicit wsgiref server
            ("--server", "wsgiref"),
        ],
    )
    def container(
        self, request: pytest.FixtureRequest, image: str
    ) -> t.Iterator[ContainerInfo]:
        """Run the pypiserver container.

        Returns the container ID.
        """
        port = get_socket()
        args = (
            "docker",
            "run",
            "--rm",
            "--publish",
            f"{port}:8080",
            "--detach",
            image,
            "run",
            "--passwords",
            ".",
            "--authenticate",
            ".",
            *request.param,
        )
        res = run(*args, capture=True)
        wait_for_container(port)
        container_id = res.out.strip()
        yield ContainerInfo(container_id, port, args)
        run("docker", "container", "rm", "-f", container_id)

    @pytest.fixture(scope="class")
    def upload_mypkg(
        self,
        container: ContainerInfo,
        mypkg_paths: t.Dict[str, Path],
    ) -> None:
        """Upload mypkg to the container."""
        run(
            sys.executable,
            "-m",
            "twine",
            "upload",
            "--repository-url",
            f"http://localhost:{container.port}",
            "--username",
            "",
            "--password",
            "",
            f"{mypkg_paths['dist_dir']}/*",
        )

    @pytest.mark.usefixtures("upload_mypkg")
    def test_download(self, container: ContainerInfo) -> None:
        """Download mypkg from the container."""
        with tempfile.TemporaryDirectory() as tmpdir:
            run(
                sys.executable,
                "-m",
                "pip",
                "download",
                "--index-url",
                f"http://localhost:{container.port}/simple",
                "--dest",
                tmpdir,
                "pypiserver_mypkg",
            )
            assert any(
                "pypiserver_mypkg" in path.name
                for path in Path(tmpdir).iterdir()
            )

    @pytest.mark.usefixtures("upload_mypkg", "cleanup")
    def test_install(self, container: ContainerInfo) -> None:
        """Install mypkg from the container.

        Note this also ensures that name normalization is working,
        since we are requesting the package name with a dash, rather
        than an underscore.
        """
        run(
            sys.executable,
            "-m",
            "pip",
            "install",
            "--force-reinstall",
            "--index-url",
            f"http://localhost:{container.port}/simple",
            TEST_DEMO_PIP_PACKAGE,
        )
        run("python", "-c", "'import pypiserver_mypkg; mypkg.pkg_name()'")

    def test_expected_server(self, container: ContainerInfo) -> None:
        """Ensure we run the server we think we're running."""
        resp = httpx.get(f"http://localhost:{container.port}")
        server = resp.headers["server"].lower()
        arg_pairs = tuple(zip(container.args, container.args[1:]))
        if (
            container.args[-1] == "pypiserver:test"
            or ("--server", "gunicorn") in arg_pairs
        ):
            # We specified no overriding args, so we should run gunicorn, or
            # we specified gunicorn in overriding args.
            assert "gunicorn" in server
        elif ("--server", "wsgiref") in arg_pairs:
            # We explicitly specified the wsgiref server
            assert "wsgiserver" in server
        elif ("--server", "waitress") in arg_pairs:
            # We explicitly specified the wsgiref server
            assert "waitress" in server
        else:
            # We overrode args, so instead of using the gunicorn default,
            # we use the `auto` option. Bottle won't choose gunicorn as an
            # auto server, so we have waitress installed in the docker container
            # as a fallback for these scenarios, since wsgiref is not production
            # ready
            assert "waitress" in server

    def test_welcome(self, container: ContainerInfo) -> None:
        """View the welcome page."""
        resp = httpx.get(f"http://localhost:{container.port}")
        assert resp.status_code == 200
        assert "pypiserver" in resp.text


class TestAuthed:
    """Test basic pypiserver functionality in a simple unauthed container."""

    HOST_PORT = get_socket()

    @pytest.fixture(scope="class")
    def container(self, image: str) -> t.Iterator[str]:
        """Run the pypiserver container.

        Returns the container ID.
        """
        with tempfile.TemporaryDirectory() as tmpdir:
            dirpath = Path(tmpdir)
            shutil.copy2(HTPASS_FILE, dirpath / "htpasswd")
            pkg_path = dirpath / "packages"
            pkg_path.mkdir(mode=0o777)

            res = run(
                "docker",
                "run",
                "--rm",
                "--publish",
                f"{self.HOST_PORT}:8080",
                "-v",
                f"{dirpath / 'htpasswd'}:/data/htpasswd",
                "--detach",
                image,
                "run",
                "--passwords",
                "/data/htpasswd",
                "--authenticate",
                "download, update",
                capture=True,
            )
            wait_for_container(self.HOST_PORT)
            container_id = res.out.strip()
            yield container_id
            run("docker", "container", "rm", "-f", container_id)

    @pytest.fixture(scope="class")
    def upload_mypkg(
        self,
        container: str,  # pylint: disable=unused-argument
        mypkg_paths: t.Dict[str, Path],
    ) -> None:
        """Upload mypkg to the container."""
        run(
            sys.executable,
            "-m",
            "twine",
            "upload",
            "--repository-url",
            f"http://localhost:{self.HOST_PORT}",
            "--username",
            "a",
            "--password",
            "a",
            f"{mypkg_paths['dist_dir']}/*",
        )

    def test_upload_failed_auth(
        self,
        container: str,  # pylint: disable=unused-argument
        mypkg_paths: t.Dict[str, Path],
    ) -> None:
        """Upload mypkg to the container."""
        run(
            sys.executable,
            "-m",
            "twine",
            "upload",
            "--repository-url",
            f"http://localhost:{self.HOST_PORT}",
            f"{mypkg_paths['dist_dir']}/*",
            check_code=lambda c: c != 0,
        )

    @pytest.mark.usefixtures("upload_mypkg")
    def test_download(self) -> None:
        """Download mypkg from the container."""
        with tempfile.TemporaryDirectory() as tmpdir:
            run(
                sys.executable,
                "-m",
                "pip",
                "download",
                "--index-url",
                f"http://a:a@localhost:{self.HOST_PORT}/simple",
                "--dest",
                tmpdir,
                "pypiserver_mypkg",
            )
            assert any(
                "pypiserver_mypkg" in path.name
                for path in Path(tmpdir).iterdir()
            )

    @pytest.mark.usefixtures("upload_mypkg")
    def test_download_failed_auth(self) -> None:
        """Download mypkg from the container."""
        with tempfile.TemporaryDirectory() as tmpdir:
            run(
                sys.executable,
                "-m",
                "pip",
                "download",
                "--index-url",
                f"http://foo:bar@localhost:{self.HOST_PORT}/simple",
                "--dest",
                tmpdir,
                "pypiserver_mypkg",
                check_code=lambda c: c != 0,
            )

    @pytest.mark.usefixtures("upload_mypkg", "cleanup")
    def test_install(self) -> None:
        """Install mypkg from the container.

        Note this also ensures that name normalization is working,
        since we are requesting the package name with a dash, rather
        than an underscore.
        """
        run(
            sys.executable,
            "-m",
            "pip",
            "install",
            "--force-reinstall",
            "--index-url",
            f"http://a:a@localhost:{self.HOST_PORT}/simple",
            TEST_DEMO_PIP_PACKAGE,
        )
        run("python", "-c", "'import pypiserver_mypkg; mypkg.pkg_name()'")

    @pytest.mark.usefixtures("upload_mypkg", "cleanup")
    def test_install_failed_auth(self) -> None:
        """Install mypkg from the container.

        Note this also ensures that name normalization is working,
        since we are requesting the package name with a dash, rather
        than an underscore.
        """
        run(
            sys.executable,
            "-m",
            "pip",
            "install",
            "--force-reinstall",
            "--no-cache",
            "--index-url",
            f"http://foo:bar@localhost:{self.HOST_PORT}/simple",
            TEST_DEMO_PIP_PACKAGE,
            check_code=lambda c: c != 0,
        )

    def test_welcome(self) -> None:
        """View the welcome page."""
        resp = httpx.get(f"http://localhost:{self.HOST_PORT}")
        assert resp.status_code == 200
        assert "pypiserver" in resp.text



from pathlib import Path

import pytest

from pypiserver.backend import listdir


def create_path(root: Path, path: Path):
    if path.is_absolute():
        raise ValueError(
            "Only test using relative paths"
            " to prevent leaking outside test environment"
        )
    fullpath = root / path
    if not fullpath.parent.exists():
        fullpath.parent.mkdir(parents=True)
    fullpath.touch()


valid_paths = ["direct-in-root.zip", "some/nested/pkg.zip"]


@pytest.mark.parametrize("path_name", valid_paths)
def test_listdir_generates_pkgfile_for_valid_package(tmp_path, path_name):
    path = Path(path_name)
    create_path(tmp_path, path)
    assert len(list(listdir(tmp_path))) == 1


invalid_paths = [
    ".hidden-pkg.zip",
    ".hidden/dir/pkg.zip",
    "in/between/.hidden/pkg.zip",
    "invalid-wheel.whl",
]


@pytest.mark.parametrize("path_name", invalid_paths)
def test_listdir_doesnt_generate_pkgfile_for_invalid_file(tmp_path, path_name):
    path = Path(path_name)
    create_path(tmp_path, path)
    assert not list(listdir(tmp_path))
